/* ************************************************************************
*  File: regen.c                                                          *
*                                                                         *
*  Usage: Contains routines to handle event based point regeneration      *
*                                                                         *
*  Written by Eric Green (ejg3@cornell.edu)                               *
************************************************************************ */

#include "conf.h"
#include "sysdep.h"


#include "structs.h"
#include "utils.h"
#include "spells.h"
#include "handler.h"
#include "events.h"


/* Player point types for events */
#define REGEN_HIT      0
#define REGEN_MANA     1
#define REGEN_MOVE     2

#define PULSES_PER_MUD_HOUR     (SECS_PER_MUD_HOUR*PASSES_PER_SEC)


/* event object structure for point regen */
struct regen_event_obj {
  struct char_data *ch;		/* character regening */
  int type;			/* HIT, MOVE, or MANA */
};


EVENTFUNC(points_event)
{
  struct regen_event_obj *regen = (struct regen_event_obj *) event_obj;
  struct char_data *ch;
  int type, gain;
  char logbuf[100];

  ch = regen->ch;
  type = regen->type;

  if (GET_POS(ch) >= POS_STUNNED) {
    /* no help for the dying */

    /*
     * Increment type of points by one.
     * If not at max, reenqueue the event.
     */

    switch (type) {
    case REGEN_HIT:
      GET_HIT(ch) = MIN(GET_HIT(ch) + 1, GET_MAX_HIT(ch));

      if (GET_POS(ch) <= POS_STUNNED)
	update_pos(ch);

      if (GET_HIT(ch) < GET_MAX_HIT(ch)) {
	/* reenqueue the event */
	gain = hit_gain(ch);
	return (PULSES_PER_MUD_HOUR / (gain ? gain : 1));
      }
      break;

    case REGEN_MANA:
      GET_MANA(ch) = MIN(GET_MANA(ch) + 1, GET_MAX_MANA(ch));
    
      if (GET_MANA(ch) < GET_MAX_MANA(ch)) {
	/* reenqueue the event */
	gain = mana_gain(ch);
	return (PULSES_PER_MUD_HOUR / (gain ? gain : 1));
      }
      break;

    case REGEN_MOVE:
      GET_MOVE(ch) = MIN(GET_MOVE(ch) + 1, GET_MAX_MOVE(ch));

      if (GET_MOVE(ch) < GET_MAX_MOVE(ch)) {
	/* reenqueue the event */
	gain = move_gain(ch);
	return (PULSES_PER_MUD_HOUR / (gain ? gain : 1));
      }
      break;

    default:
      sprintf(logbuf, "SYSERR:  Unknown points event type %d", type);
      log(logbuf);
      free(event_obj);
      return 0;
    }
  }

  /* kill this event */
  GET_POINTS_EVENT(ch, type) = NULL;
  free(event_obj);
  return 0;
}


/*
 * subtracts amount of hitpoints from ch's current and starts points event
 */
void alter_hit(struct char_data *ch, int amount)
{
  struct regen_event_obj *regen;
  long time;
  int gain;

  GET_HIT(ch) = MIN(GET_HIT(ch) - amount, GET_MAX_HIT(ch));

  if (GET_HIT(ch) <= HIT_INCAP)
    return;

  if (GET_HIT(ch) < GET_MAX_HIT(ch) && !GET_POINTS_EVENT(ch, REGEN_HIT)) {
    CREATE(regen, struct regen_event_obj, 1);	
    regen->ch = ch;
    regen->type = REGEN_HIT;
    gain = hit_gain(ch);
    time = PULSES_PER_MUD_HOUR / (gain ? gain : 1);
    GET_POINTS_EVENT(ch, REGEN_HIT) = event_create(points_event, regen, time);

    if (amount >= 0) {
	
      /*
       * if the character gained hp, update position and
       * restart mana and move regeneration if needed.
       */
      update_pos(ch);
      alter_mana(ch, 0);
      alter_move(ch, 0);
    }
  }
}


/*
 * subtracts amount of mana from ch's current and starts points event
 */
void alter_mana(struct char_data *ch, int amount)
{
  struct regen_event_obj *regen;
  long time;
  int gain;

  GET_MANA(ch) = MIN(GET_MANA(ch) - amount, GET_MAX_MANA(ch));
 
  if (!GET_POINTS_EVENT(ch, REGEN_MANA) && (GET_MANA(ch) < GET_MAX_MANA(ch))) {

    /* make sure the character isn't dying */
    if (GET_POS(ch) >= POS_STUNNED) {
      CREATE(regen, struct regen_event_obj, 1);
      regen->ch = ch;
      regen->type = REGEN_MANA;
      gain = mana_gain(ch);
      time = PULSES_PER_MUD_HOUR / (gain ? gain : 1);
      GET_POINTS_EVENT(ch, REGEN_MANA) =
	event_create(points_event, regen, time);
    }
  }
}


/*
 * subtracts amount of moves from ch's current and starts points event
 */
void alter_move(struct char_data *ch, int amount)
{
  struct regen_event_obj *regen;
  long time;
  int gain;

  GET_MOVE(ch) = MIN(GET_MOVE(ch) - amount, GET_MAX_MOVE(ch));
 
  if (!GET_POINTS_EVENT(ch, REGEN_MOVE) && (GET_MOVE(ch) < GET_MAX_MOVE(ch))) {

    /* make sure the character isn't dying */
    if (GET_POS(ch) >= POS_STUNNED) {
      CREATE(regen, struct regen_event_obj, 1);
      regen->ch = ch;
      regen->type = REGEN_MOVE;
      gain = move_gain(ch);
      time = PULSES_PER_MUD_HOUR / (gain ? gain : 1);
      GET_POINTS_EVENT(ch, REGEN_MOVE) =
	event_create(points_event, regen, time);
    }
  }
}


/* updates regen rates.  Use when big regen rate changes are made */
void check_regen_rates(struct char_data *ch)
{
  struct regen_event_obj *regen;
  int type, gain = 0;
  long time;

  if (GET_HIT(ch) <= HIT_INCAP)
    return;

  for (type = REGEN_HIT; type <= REGEN_MOVE; type++) {
    
    switch (type) {
    case REGEN_HIT:
      if (GET_HIT(ch) >= GET_MAX_HIT(ch))
	continue;
      gain = hit_gain(ch);
      break;

    case REGEN_MANA:
      if (GET_MANA(ch) >= GET_MAX_MANA(ch))
	continue;
      gain = mana_gain(ch);
      break;

    case REGEN_MOVE:
      if (GET_MOVE(ch) >= GET_MAX_MOVE(ch))
	continue;
      gain = move_gain(ch);
      break;
    }

    time = PULSES_PER_MUD_HOUR / (gain ? gain : 1);

    if (!GET_POINTS_EVENT(ch, type) ||
	(time < event_time(GET_POINTS_EVENT(ch, type)))) {

      /* take off old event, create updated event */
      if (GET_POINTS_EVENT(ch, type))
	event_cancel(GET_POINTS_EVENT(ch, type));
      CREATE(regen, struct regen_event_obj, 1);
      regen->ch = ch;
      regen->type = type;
      GET_POINTS_EVENT(ch, type) = event_create(points_event, regen, time);
    }
  }
}
